/* TD10. Plus courts chemins */

import java.lang.reflect.Field;
import java.util.PriorityQueue;

// Algorithme de Dijkstra
class Dijkstra {
	final Graph g; // le graphe de travail
	final int source; // source du plus court chemin recherche
	final int dest; // destination du plus court chemin recherche
	private Fenetre f; // fenetre pour la visualisation

	// Questions 1.1, 1.2 et 4
	
	// @OFF
	int dist[]; // tableau des distances intermédiaires
	int pred[]; // tableau des prédécesseurs
	boolean settled[]; 

	int steps;

	PriorityQueue<Node> unsettled;
	// @ON
	
	/* Méthodes à compléter */
	
	// Questions 1.1 et 1.2
	
	// constructeur
	Dijkstra(Graph g, int source, int dest) {
		this.g = g;
		this.source = source;
		this.dest = dest;
		
		// @OFF
		dist = new int[g.nbVertices];
		pred = new int[g.nbVertices];
		settled = new boolean[g.nbVertices];
		
		for (int i = 0; i < g.nbVertices; i++) {
			dist[i] = Integer.MAX_VALUE;
			pred[i] = -1;
			// Cette ligne n'est pas nécessaire car les booléens sont initialisé à false par défaut
			settled[i] = false; 
		}
		
		dist[source] = 0;
		pred[source] = source;
		
		unsettled = new PriorityQueue<>();
		unsettled.add(new Node(source, 0));
		
		steps = 0;
		/* @ON
		throw new Error("Constructeur Dijkstra(Graph, int, int) à compléter (Questions 1.1 et 1.2)");		
		@OFF @ON */
	}
	
	// Question 2.1 et 2.2

	// mise a jour de la distance, de la priorite, et du predecesseur d'un sommet
	void update(int succ, int current) {
		// @OFF
		int distByCurrent = dist[current] + g.weight(current, succ);
		
		if (distByCurrent < dist[succ]) {
			dist[succ] = distByCurrent;
			pred[succ] = current;
			unsettled.add(new Node(succ, distByCurrent));
			g.drawUnsettledPoint(f, succ);
		}
		/* @ON
		throw new Error("Méthode Dijkstra.update(int, int) à implémenter (Question 2.1)");
		@OFF @ON */
	}
	
	// Question 2.1

	// trouve le prochain sommet de unsettled non traite
	int nextNode() {
		// @OFF
		while (!unsettled.isEmpty()) {
			Node node = unsettled.poll();
			if (!settled[node.id]) 
				return node.id;
		}
	
		return -1;
		/* @ON
		throw new Error("Méthode Dijkstra.nextNode() à implémenter (Question 2.1)");		
		@OFF @ON */
	}
	
	// Questions 2.1, 2.2 et 4

	// une etape de l'algorithme Dijkstra
	int oneStep() {
		// @OFF
		slow();
		steps++;
		
		int current = nextNode();
		if (current == -1)
			return -1;
		
		settled[current] = true;
		for (int next : g.successors(current)) {
			update(next, current);
		}
		
		g.drawSettledPoint(f, current);
		return current;
		/* @ON
		throw new Error("Méthode Dijkstra.oneStep() à implémenter (Question 2.1)");		
		@OFF @ON */
	}
	
	// Question 2.1

	// algorithme de Dijkstra complet
	int compute() {
		// @OFF
		int current;
		
		do {
			current = oneStep();
		} while (current != dest && current != -1);

		return (current == dest) ? dist[current] : -1;
		/* @ON
		throw new Error("Méthode Dijkstra.compute() à implémenter (Question 2.1)");				
		@OFF @ON */
	}
	
	// Question 4
	
	public int getSteps() {
		// @OFF
		return steps;		
		/* @ON
		throw new Error("Méthode Dijkstra.getSteps() à implémenter (Question 4)");
		@OFF @ON */
				
	}
	
	/* Méthodes à ne pas changer */
	
	// ralentisseur visualisation
	void slow(){
		if(f == null) return;
		try {
			Thread.sleep(5);
		} catch (InterruptedException e) {}
	}
	
	void setFenetre (Fenetre f) { this.f = f; }

	// Cette fonction vérifie si le vecteur 'int[] name' est 
	// présent dans la classe et le renvoie. Sinon, renvoie null 
	private int[] getIntArray(String name) {
		Field field = null;
		for (Field f : getClass().getDeclaredFields()) {
			if (f.getName().equals(name)) {
				field = f;
				break;
			}
		}
		if (field == null)
			return null;
		
		int[] result = null;
		try {
			result = (int[]) field.get(this);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		}
		return result;
	}
	
	public int[] getPred() {
		return getIntArray("pred");
	}
	
	public int[] getDist() {
		return getIntArray("dist");
	}	
		
	public void draw () {
		g.drawSourceDestination(f, source, dest);
		g.drawPath(f, getPred(), dest);
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof Dijkstra))
			return false;
		
		Dijkstra that = (Dijkstra) obj;
		return g.equals(that.g) && source == that.source && dest == that.dest;
	}
}
